@@ -8,8 +8,8 @@
 
    This code is free for all uses.
 
-   $Revision: 1.48 $
-   $Date: 2005/08/28 18:54:50 $
+   $Revision: 1.55 $
+   $Date: 2005/11/24 01:29:21 $
  */
 
 #include <stdio.h>
@@ -56,6 +56,7 @@ ushort sget2 (uchar *s)
   else				/* "MM" means big-endian */
     return s[0] << 8 | s[1];
 }
+#define sget2(s) sget2((uchar *)s)
 
 ushort get2()
 {
@@ -71,6 +72,7 @@ int sget4 (uchar *s)
   else
     return s[0] << 24 | s[1] << 16 | s[2] << 8 | s[3];
 }
+#define sget4(s) sget4((uchar *)s)
 
 int get4()
 {
@@ -181,7 +183,7 @@ void nikon_decrypt (uchar ci, uchar cj, int tag, int i, int size, uchar *buf)
     0xc6,0x67,0x4a,0xf5,0xa5,0x12,0x65,0x7e,0xb0,0xdf,0xaf,0x4e,0xb3,0x61,0x7f,0x2f } };
   uchar ck=0x60;
 
-  if (strncmp (buf, "02", 2)) return;
+  if (strncmp ((char *)buf, "02", 2)) return;
   ci = xlat[0][ci];
   cj = xlat[1][cj];
   printf("Decrypted tag 0x%x:\n%*s", tag, (i & 31)*3, "");
@@ -200,7 +202,6 @@ void nef_parse_makernote (base)
   short sorder;
   char buf[10];
 
-  puts("  Nikon MakerNote:");
 /*
    The MakerNote might have its own TIFF header (possibly with
    its own byte-order!), or it might just be a table.
@@ -214,11 +215,13 @@ void nef_parse_makernote (base)
     offset = get4();
     fseek (ifp, offset-8, SEEK_CUR);
   } else if (!strncmp (buf,"FUJIFILM",8) ||
+	     !strncmp (buf,"SONY",4) ||
 	     !strcmp  (buf,"Panasonic")) {
     order = 0x4949;
     fseek (ifp,  2, SEEK_CUR);
   } else if (!strcmp (buf,"OLYMP") ||
 	     !strcmp (buf,"LEICA") ||
+	     !strcmp (buf,"Ricoh") ||
 	     !strcmp (buf,"EPSON"))
     fseek (ifp, -2, SEEK_CUR);
   else if (!strcmp (buf,"AOC"))
@@ -228,6 +231,7 @@ void nef_parse_makernote (base)
 
   entries = get2();
   if (entries > 100) return;
+  puts("  MakerNote:");
   while (entries--) {
     save = ftell(ifp);
     tag  = get2();
@@ -275,7 +279,7 @@ void nef_parse_makernote (base)
     fseek (ifp, save+12, SEEK_SET);
   }
   nikon_decrypt (serial, key, 0x91, 4, sizeof buf91, buf91);
-  if (!strncmp (buf97, "0205", 4))
+  if (!strncmp ((char *)buf97, "0205", 4))
     nikon_decrypt (serial, key, 0x97, 4, 284, buf97);
   else
     nikon_decrypt (serial, key, 0x97, 284, sizeof buf97, buf97);
@@ -303,10 +307,29 @@ void nef_parse_exif(int base)
 
 void parse_mos(int level);
 
+void sony_decrypt (unsigned *data, int len, int start, int key)
+{
+  static unsigned pad[128], p;
+
+  if (start) {
+    for (p=0; p < 4; p++)
+      pad[p] = key = key * 48828125 + 1;
+    pad[3] = pad[3] << 1 | (pad[0]^pad[2]) >> 31;
+    for (p=4; p < 127; p++)
+      pad[p] = (pad[p-4]^pad[p-2]) << 1 | (pad[p-3]^pad[p-1]) >> 31;
+    for (p=0; p < 127; p++)
+      pad[p] = htonl(pad[p]);
+  }
+  while (len--)
+    *data++ ^= pad[p++ & 127] = pad[(p+1) & 127] ^ pad[(p+65) & 127];
+}
+
 int parse_tiff_ifd (int base, int level)
 {
   int entries, tag, type, count, slen, save, save2, val, i;
   int comp=0;
+  unsigned *buf, sony_offset=0, sony_length=0, sony_key=0;
+  FILE *sfp;
 
   entries = get2();
   if (entries > 255) return 1;
@@ -382,13 +405,18 @@ int parse_tiff_ifd (int base, int level)
 	}
 	break;
       case 0x201:
-	if (strncmp(make,"OLYMPUS",7) || !thumb_offset)
+	if (!thumb_offset || (strncmp(make,"OLYMPUS",7)
+			   && strncmp(make,"SONY",4)))
 	  thumb_offset = val;
 	break;
       case 0x202:
-	if (strncmp(make,"OLYMPUS",7) || !thumb_length)
+	if (!thumb_length || (strncmp(make,"OLYMPUS",7)
+			   && strncmp(make,"SONY",4)))
 	  thumb_length = val;
 	break;
+      case 29184: sony_offset = val;  break;
+      case 29185: sony_length = val;  break;
+      case 29217: sony_key = get4();  break;
       case 34310:
 	parse_mos(0);
 	break;
@@ -398,10 +426,31 @@ int parse_tiff_ifd (int base, int level)
 	break;
       case 50706:
 	is_dng = 1;
+	break;
+      case 50740:
+	if (count != 4 || type != 1) break;
+	puts("Sony SR2 private IFD:");
+	fseek (ifp, get4()+base, SEEK_SET);
+	parse_tiff_ifd (base, level+1);
     }
 cont:
     fseek (ifp, save+12, SEEK_SET);
   }
+  if (sony_length && (buf = malloc(sony_length))) {
+    fseek (ifp, sony_offset, SEEK_SET);
+    fread (buf, sony_length, 1, ifp);
+    sony_decrypt (buf, sony_length/4, 1, sony_key);
+    sfp = ifp;
+    if ((ifp = tmpfile())) {
+      fwrite (buf, sony_length, 1, ifp);
+      fseek (ifp, 0, SEEK_SET);
+      puts ("Sony SR2 encrypted IFD:");
+      parse_tiff_ifd (-sony_offset, level);
+      fclose (ifp);
+    }
+    ifp = sfp;
+    free (buf);
+  }
   if (((comp == 6 && !strcmp(make,"Canon")) ||
        (comp == 7 && is_dng)) && offset) {
     thumb_offset = offset;
@@ -575,9 +624,39 @@ int parse_jpeg (int offset)
   return 1;
 }
 
+void parse_riff (int level)
+{
+  unsigned i, size, end, save;
+  char tag[4], type[4], buf[64];
+
+  order = 0x4949;
+  fread (tag, 4, 1, ifp);
+  size = get4();
+  if (isdigit(tag[0])) {
+    fseek (ifp, size, SEEK_CUR);
+    return;
+  }
+  printf ("%*.4s size %d", level*4+4, tag, size);
+  if (!memcmp(tag,"RIFF",4) || !memcmp(tag,"LIST",4)) {
+    end = ftell(ifp) + size;
+    fread (type, 4, 1, ifp);
+    printf (" type %.4s:\n", type);
+    while (ftell(ifp) < end)
+      parse_riff (level+1);
+  } else {
+    save = ftell(ifp);
+    fread (buf, 1, 40, ifp);
+    printf (": ");
+    for (i=0; i < 40 && isprint(buf[i]); i++)
+      putchar (buf[i]);
+    putchar ('\n');
+    fseek (ifp, save+size, SEEK_SET);
+  }
+}
+
 void parse_mos(int level)
 {
-  uchar data[256];
+  char data[256];
   int i, skip, save;
   char *cp;
 
@@ -972,9 +1051,11 @@ void parse_fuji (int offset)
   int entries, tag, len;
 
   fseek (ifp, offset, SEEK_SET);
-  fseek (ifp, get4(), SEEK_SET);
+  if (!(len = get4())) return;
+  printf ("Fuji table at %d:\n",len);
+  fseek (ifp, len, SEEK_SET);
   entries = get4();
-  if (entries > 60) return;
+  if (entries > 255) return;
   while (entries--) {
     tag = get2();
     len = get2();
@@ -1063,9 +1144,9 @@ int identify()
       fseek (ifp, hlen, SEEK_SET);
     } else
       parse_tiff (0);
-  } else if (!memcmp (head, "\0MRM", 4)) {
+  } else if (!memcmp (head,"\0MRM",4)) {
     parse_minolta();
-  } else if (!memcmp (head, "FUJIFILM", 8)) {
+  } else if (!memcmp (head,"FUJIFILM",8)) {
     fseek (ifp, 84, SEEK_SET);
     toff = get4();
     tlen = get4();
@@ -1074,9 +1155,12 @@ int identify()
     parse_tiff (toff+12);
     thumb_offset = toff;
     thumb_length = tlen;
-  } else if (!memcmp (head, "DSC-Image", 9))
+  } else if (!memcmp (head,"RIFF",4)) {
+    fseek (ifp, 0, SEEK_SET);
+    parse_riff(0);
+  } else if (!memcmp (head,"DSC-Image",9))
     parse_rollei();
-  else if (!memcmp (head, "FOVb", 4))
+  else if (!memcmp (head,"FOVb",4))
     parse_foveon();
   fseek (ifp, 8, SEEK_SET);
   parse_mos(0);
